/**
   @file
   @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_UTIL_MULTI_SEQ_H_INCLUDED
#define CNOID_UTIL_MULTI_SEQ_H_INCLUDED

#include <Eigen/StdVector>
#include "AbstractSeq.h"
#include "Array2D.h"
#include <algorithm>

namespace cnoid {

    template <typename ElementType, typename Allocator = std::allocator<ElementType> >
    class MultiSeq : public Array2D<ElementType, Allocator>, public AbstractMultiSeq
    {
        typedef MultiSeq<ElementType, Allocator> MultiSeqType;
        
    public:
        typedef Array2D<ElementType, Allocator> Array2DType;
        typedef typename Array2DType::Element Element;
        typedef boost::shared_ptr< MultiSeqType > Ptr;
        typedef typename Array2DType::Row Frame;
        typedef typename Array2DType::Column Part;

        MultiSeq(const char* seqType, int numParts = 1, int numFrames = 0, double frameRate = 100.0)
            : AbstractMultiSeq(seqType),
              Array2DType(numFrames, numParts) {
            frameRate_ = frameRate;
        }

        MultiSeq(const MultiSeqType& org)
            : AbstractMultiSeq(org),
              Array2DType(org) {
            frameRate_ = org.frameRate_;
        }
    
        virtual ~MultiSeq() { }

        MultiSeqType& operator=(const MultiSeqType& rhs) {
            if(this != &rhs){
                AbstractMultiSeq::operator=(rhs);
                Array2DType::operator=(rhs);
                frameRate_ = rhs.frameRate_;
            }
            return *this;
        }

        virtual void setDimension(int newNumFrames, int newNumParts, bool clearNewElements = false) {

            const int prevNumParts = numParts();
            const int prevNumFrames = numFrames();

            Array2DType::resize(newNumFrames, newNumParts);

            if(clearNewElements){
                if(newNumParts == prevNumParts){
                    if(newNumFrames > prevNumFrames){
                        std::fill(Array2DType::begin() + prevNumFrames * newNumParts,
                                  Array2DType::end(),
                                  defaultValue());
                    }
                } else {
                    std::fill(Array2DType::begin(), Array2DType::end(), defaultValue());
                }
            }
        }

        virtual double getFrameRate() const {
            return frameRate_;
        }

        inline double frameRate() const {
            return frameRate_;
        }

        virtual void setFrameRate(double frameRate) {
            frameRate_ = frameRate;
        }

        virtual void setNumParts(int newNumParts, bool clearNewElements = false) {
            setDimension(numFrames(), newNumParts, clearNewElements);
        }

        virtual int getNumFrames() const {
            return Array2DType::rowSize();
        }

        inline int numFrames() const {
            return Array2DType::rowSize();
        }

        virtual void setNumFrames(int newNumFrames, bool clearNewElements = false) {
            setDimension(newNumFrames, numParts(), clearNewElements);
        }

        virtual int getNumParts() const {
            return Array2DType::colSize();
        }

        inline int numParts() const {
            return Array2DType::colSize();
        }

        /**
           @if jp
           シーケンスの時間長を返す。
           @note この時間 *未満* の時間については有効なデータが存在する。
           この時間のデータは存在しないので、そのようなアクセスしてはいけない。
           @endif
        */
        inline double timeLength() const {
            return numFrames() / frameRate();
        }

        inline int frameOfTime(double time) const {
            return (int)(time * frameRate_);
        }
            
        inline double timeOfFrame(int frame) const {
            return (frame / frameRate_);
        }
            
        inline Part part(int index) {
            return Array2DType::column(index);
        }

        inline Frame frame(int index) {
            return Array2DType::row(index);
        }

    protected:

        double frameRate_;

        virtual ElementType defaultValue() const { return ElementType(); }
    };
}

#endif
